function Paper_LRBFGS_VS_TFOCS_noise(kappa)
% Reproduce the experiments in Wen Huang's Dissertation
% Test the performance of Riemannian algorithms for the Brockett problem on
% the Stiefel manifold.
% 
% By Wen Huang

    ps = [1, 2, 4, 8, 16];
    tab = zeros(6, length(ps) + 1);
    n1 = 64;
    n2 = 64;
%     kappa = 1/ sqrt(n1 * n2);
%     kappa = 1e-6;
    tau = 1e-5;
    
    [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp, ave_time, mse] = test_PhaseRetrieval(n1, n2, tau, 1/ sqrt(n1 * n2));
    tab(1, 1) = length(F);
    tab(2, 1) = nf;
    tab(3, 1) = ng;
    tab(4, 1) = F(end);
    tab(5, 1) = mse;
    tab(6, 1) = ave_time;

%     kappa = 0;
    for i = 1 : length(ps)
        [iter, nf, ng, f_final, mse, ave_time] = test_TFOCS_PhaseRetrieval(n1, n2, ps(i), tau, kappa);
        tab(1, i + 1) = iter;
        tab(2, i + 1) = nf;
        tab(3, i + 1) = ng;
        tab(4, i + 1) = f_final;
        tab(5, i + 1) = mse;
        tab(6, i + 1) = ave_time;
    end
    
    fout = fopen(['Paper_LRBFGS_VS_TFOCS_tab_' num2str(tau) '_' num2str(kappa) '.txt'],'w');
    for i = 1 : size(tab, 1)
        for j = 1 : size(tab, 2)
            if(mod(tab(i, j), 1) == 0)
                if(j < size(tab, 2))
                    fprintf(fout,'%d & ',tab(i, j));
                else
                    fprintf(fout,'%d',tab(i, j));
                end
            else
                tabstr = outputfloat(tab(i, j));
                if(j < size(tab, 2))
                    fprintf(fout,['$' tabstr '$ & ']);
                else
                    fprintf(fout,['$' tabstr '$']);
                end
            end
        end
        fprintf(fout,' \\\\\n');
    end
    fclose(fout);
end

function str = outputfloat(x)
    if(x <= 0)
        str = '';
        return;
    end
    p = log(x)/log(10);
    p = - ceil(-p);
    x = round(x * 10^(-p) * 100);
    x = x / 100;
    strx = sprintf('%3.2f', x);
    if(p ~= 0)
        str = [strx '_{' num2str(p) '}'];
    else
        str = strx;
    end
end

function [X, F, G, T, timecost, nf, ng, nR, nH, nV, nVp, ave_time, mse] = test_PhaseRetrieval(n1, n2, tau, kappa)
% test file for test_PhaseRetrieval.
% 
% By Wen Huang
    fprintf('test for Phase Retrieval problem. \n');
    r = 1;
%     r = 943
    rand('state', r);
    randn('state', r);
%     n1 = 64;
%     n2 = 64;
    global count
    count = 0;
    n = n1 * n2;
    l = 6;
    m = l * n;
    k = 4;
    method = 3;
    
    fprintf('In this case, n1 = %d, n2 = %d, n = %d, l = %d. b, and x0 are chosen randomly. Random seed is %d \n', n1, n2, n, l, r);
    
    metric = 1;

    %% generate masks and a true solution.
    global iscomplex
    iscomplex = 1;
    global isfixedrank
    isfixedrank = 0;
    masks = complex(randn(n1, n2, l), iscomplex * randn(n1, n2, l));
    phase = complex(randn(n1, n2), iscomplex * randn(n1, n2));
    phase = phase / norm(phase);
    b = zeros(m, 1);
    btrue = zeros(m, 1);
    for i = 1 : l
        zi = reshape(fft2(phase .* masks(:, :, i)) / sqrt(n1 * n2), [], 1);
        btrue((i - 1) * n + 1 : i * n) = conj(zi) .* zi;
        b((i - 1) * n + 1 : i * n) = conj(zi) .* zi + tau * randn(n, 1);
    end
    b2n = norm(b)^2
    SNR = 10 * log(norm(btrue) / norm(btrue - b)) / log(10)
%     bnorm = norm(b)
%     phasenorm = norm(phase, 'fro')
    
    %%
    initialtype = 3; % 0: random, 1: true solution, 2: modification of true solution
    x0eps = 0.1;
    if((initialtype == 1 || initialtype == 2) && k ~= 1)
        fprintf('k should be 1\n');
        return;
    end
    
    if(initialtype == 0)
        [x1.U, r] = qr(complex(randn(n, k), iscomplex * randn(n, k)), 0);
%         x1.D = spdiags(rand(k, 1), 0, k, k);
        x1.D = spdiags(ones(k, 1), 0, k, k);
    elseif(initialtype == 1)
        x1.U = reshape(phase, [], 1) / norm(reshape(phase, [], 1));
        x1.D = norm(reshape(phase, [], 1))^2;
    elseif(initialtype == 2)
        temp = phase + randn(n1, n2) * x0eps;
        x1.U = reshape(temp, [], 1) / norm(reshape(temp, [], 1));
        x1.D = norm(reshape(temp, [], 1))^2;
    elseif(initialtype == 3)
        tic
        x1.U = randn(n, k);
        x1.U = orth(x1.U);;
        for i = 1 : 20
            x1.U = Ax(x1.U, masks, b, kappa);
            x1.U = orth(x1.U);;
        end
        x1.D = eye(k);
        toc
    end
    
    if(metric == 1 || metric == 2)
        x1.U = full(x1.U * sqrt(x1.D));
        if(method == 2 || method == 5)
            params.H0 = eye(n * k);
            params.B0 = eye(n * k);
        end
    end
    
    params.x0 = x1;
    params.error = 1e-8;
    params.max_t = inf;
    params.debug = 0;
    params.manifold = 2;
    params.m = 4;
    params.alpha = 1e-4;
    params.beta = 0.999;
    params.min_Delta = 1e-8;
    params.minstepsize = eps;
    params.maxstepsize = 200;
%     params.kappa = 0.9;
%     params.retraction = Retraction;
%     params.vector_transport = VT;
    params.Delta_bar = inf;
    params.Delta0 = 1;
    params.StopCriterion = 0;
    params.restart = 0;
    if(method == 11)
        params.linesearch = 1;
    end
    times = 0;
    total_timecost = 0;
    while(total_timecost < 60 && times < 5)
        fprintf('%d: ', times);
        [X, F, G, T1, timecost, nf, ng, nR, nH, nV, nVp] = driver_PhaseRetrieval(b, masks, method, params, phase, metric, 0.9, 0, kappa);
        if(times == 0)
            T = T1;
        else
            T = T + T1;
        end
        total_timecost = total_timecost + timecost;
        times = times + 1;
    end
    if(method == 13)
        return;
    end
    mse = MSE_error(X{end}.U, reshape(phase, [], 1));
    T = T / times;
    ave_time = total_timecost / times;
    fprintf('average time cost of %d times is %f. \n', times, total_timecost / times);
end

function output = Ax(U, masks, b, kappa)
    x.U = U;
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    m = l * n;
    k = size(x.U, 2);
    ZY = zeros(m, k);
    for i = 1 : l
        for j = 1 : k
            temp = fft2(reshape(x.U(:, j), n1, n2) .* masks(:, :, i)) / sqrt(n1 * n2);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
        end
    end
    DZY = zeros(m, k);
    for i = 1 : k
        DZY(:, i) = b .* ZY(:, i);
    end
    %% Z^H *DZY
    Y = zeros(n, k);
    for i = 1 : l
        temp = zeros(n, k);
        for j = 1 : k
            temp(:, j) = reshape(ifft2(reshape(DZY((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
            temp(:, j) = conj(reshape(masks(:, :, i), [], 1)) .* temp(:, j);
        end
        Y = Y + temp;
    end
    output = Y;
end

function mse = MSE_error(x, phase)
    X1 = [real(x), imag(x)];
    X2 = [real(phase), imag(phase)];
    [U, S, V] = svd(X2'*X1);
    O = V * U';
    mse = norm(X1 * O - X2, 'fro') / norm(X2, 'fro');
end

function [iter, nf, ng, f_final, mse, ave_time] = test_TFOCS_PhaseRetrieval(n1, n2, p, tau, kappa)
%% Now, add in constraints
    r = rand();
%     r = floor(r * 10000)
    r = 1;
    rand('state', r);
    randn('state', r);
%     n1 = 64;
%     n2 = 64;
    n = n1 * n2;
    l = 6;
    m = l * n;
    k = p;

    %% generate masks and a true solution.
    global iscomplex
    iscomplex = 1;
    masks = complex(randn(n1, n2, l), iscomplex * randn(n1, n2, l));
    phase = complex(randn(n1, n2), iscomplex * randn(n1, n2));
    phase = phase / norm(phase);
    b = zeros(m, 1);
    for i = 1 : l
        zi = reshape(fft2(phase .* masks(:, :, i)) / sqrt(n1 * n2), [], 1);
        b((i - 1) * n + 1 : i * n) = conj(zi) .* zi + tau * randn(n, 1);
    end
%     bnorm = norm(b)
%     phasenorm = norm(phase, 'fro')
    
    initialtype = 3; % 0: random, 1: true solution, 2: modification of true solution
    x0eps = 0.1;
    if((initialtype == 1 || initialtype == 2) && k ~= 1)
        fprintf('k should be 1\n');
        return;
    end
    
    if(initialtype == 0)
        [x1.U, r] = qr(complex(randn(n, k), iscomplex * randn(n, k)), 0);
%         x1.D = spdiags(rand(k, 1), 0, k, k);
        x1.D = spdiags(ones(k, 1), 0, k, k);
    elseif(initialtype == 1)
        x1.U = reshape(phase, [], 1) / norm(reshape(phase, [], 1));
        x1.D = norm(reshape(phase, [], 1))^2;
    elseif(initialtype == 2)
        temp = phase + randn(n1, n2) * x0eps;
        x1.U = reshape(temp, [], 1) / norm(reshape(temp, [], 1));
        x1.D = norm(reshape(temp, [], 1))^2;
    elseif(initialtype == 3)
        tic
        x1.U = randn(n, k);
        x1.U = orth(x1.U);;
        for i = 1 : 20
            x1.U = Ax(x1.U, masks, b, kappa);
            x1.U = orth(x1.U);;
        end
        x1.D = eye(k);
        toc
    end
    x0 = full(x1.U * sqrt(x1.D));

    smoothF = @(x)f_gf(x, masks, b, kappa);
    
    opts = [];
    opts.tol        = 1e-6;
    opts.maxits     = 2000;
    opts.alg = 'N83_PR';
%     opts.stopCrit = 2;
%     opts.stopFcn = @stopFcn;
%     opts.err = 1e-6;
    
    times = 0;
    total_timecost = 0;
    while(total_timecost < 60 && times < inf)
        fprintf('times:%d, initial p:%d \n', times, k);
        [ x, out, optsOut ] = tfocs(smoothF, [], [], x0, opts );
        timecost = optsOut.timecost;
        total_timecost = total_timecost + timecost;
        times = times + 1;
    end
    ave_time = total_timecost / times;
    iter = optsOut.iter;
    nf = optsOut.nf;
    ng = optsOut.ng;
    f_final = f_gf(x(:, 1), masks, b, kappa);
    mse = MSE_error(x(:, 1), reshape(phase, [], 1));
return;
end

function stop = stopFcn(f, x)
    stop = 0;% f < 1e-10;
return;
end

function [f, gf] = f_gf(x, masks, b, kappa)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    k = size(x, 2);
    m = l * n;
    D = zeros(m, 1);
    ZY = zeros(m, k);
    for i = 1 : l
        sZqi = zeros(n, 1);
        for j = 1 : k
            temp = reshape(fft2(reshape(x(:, j), n1, n2) / sqrt(n1 * n2) .* masks(:, :, i)), [], 1);
            ZY((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
            sZqi = sZqi + conj(temp) .* temp;
        end
        D((i - 1) * n + 1 : i * n) = reshape(sZqi, [], 1) - b((i - 1) * n + 1 : i * n);
    end
    f = norm(D)^2 / norm(b, 'fro')^2;
    f = f + kappa * norm(x, 'fro')^2;%% add penelty term
    gf = @(v)gfv(v, D, masks, b, kappa);
return
end

function output = gfv(v, D, masks, b, kappa)
    [n1, n2, l] = size(masks);
    n = n1 * n2;
    s = size(v, 2);
    m = length(D);
    Zv = zeros(m, s);    
    for i = 1 : l
        for j = 1 : s
            temp = reshape(fft2(reshape(v(:, j), n1, n2) / sqrt(n1 * n2) .* masks(:, :, i)), [], 1);
            Zv((i - 1) * n + 1 : i * n, j) = reshape(temp, [], 1);
        end
    end
    
    DZv = spdiags(D, 0, m, m) * Zv;
    %% Z^H *DZv
    Y = zeros(n, s);
    for i = 1 : l
        temp = zeros(n, s);
        for j = 1 : s
            temp(:, j) = reshape(ifft2(reshape(DZv((i - 1) * n + 1 : i * n, j) * sqrt(n1 * n2), n1, n2)), [], 1);
        end
        temp = conj(spdiags(reshape(masks(:, :, i), [], 1), 0, n, n)) * temp;
        Y = Y + temp;
    end
    output = 2 * Y / norm(b, 'fro')^2 + kappa * v;
return
end

